#include <WiFi.h>
#include <WiFiClientSecure.h>
#include <PubSubClient.h>
#include <ArduinoJson.h>
#include <vector>
#include <map>

#define M_DOT 100
#define M_DASH 300
#define M_WORD 700

const char* ssid = "BS-GAST";
const char* password = "bs2020##";

const char* mqtt_server = "emqx.kubeberry.com";
const int mqtt_port = 8883;
const char* mqtt_user = "mqtt-user";
const char* mqtt_pass = "dieses passwort ist nicht korrekt";

WiFiClientSecure secureClient;
PubSubClient client(secureClient);

std::vector<int> listeningPins;
std::map<int, int> lastValues;
unsigned long lastReadTime = 0;
const unsigned long readInterval = 100;

void setup_wifi() {
  delay(10);
  Serial.print("Connecting to ");
  Serial.println(ssid);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("\nWiFi connected");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
}

void addListeningPin(int pin) {
  if (std::find(listeningPins.begin(), listeningPins.end(), pin) == listeningPins.end()) {
    pinMode(pin, INPUT);
    listeningPins.push_back(pin);
    Serial.printf("Now listening to pin %d\n", pin);
  } else {
    Serial.printf("Already listening to pin %d\n", pin);
  }
}

void dot(int pin) {
  ledcWrite(pin, 440);
  delay(1000);
  ledcWrite(pin, 0);
}

void useBuzzer(int pin) {
  ledcAttach(pin, 1000, 10);

  dot(pin);

  ledcDetach(pin);
}

void callback(char* topic, byte* payload, unsigned int length) {
  String msg;
  for (unsigned int i = 0; i < length; i++) {
    msg += (char)payload[i];
  }
  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.print("]: ");
  Serial.println(msg);

  StaticJsonDocument<256> doc;
  DeserializationError error = deserializeJson(doc, msg);
  if (error) {
    Serial.print("JSON parse error: ");
    Serial.println(error.c_str());
    return;
  }

  int pin = doc["pin"];
  const char* mode = doc["mode"];
  int state = doc["state"] | -1;

  if (pin < 0 || pin > 39) {
    Serial.println("Invalid pin");
    return;
  }

  if (mode && strcmp(mode, "listen") == 0) {
    addListeningPin(pin);
  } else if (mode && strcmp(mode, "buzzer") == 0) {
    useBuzzer(pin);
  } else if (state != -1) {
    pinMode(pin, OUTPUT);
    digitalWrite(pin, state == 0 ? LOW : HIGH);
    Serial.printf("Set pin %d to %s\n", pin, state == 0 ? "LOW" : "HIGH");
  } else {
    Serial.println("Unknown command");
  }
}

void reconnect() {
  while (!client.connected()) {
    Serial.print("Connecting to MQTT...");
    String clientId = "ESP32Client-" + String(random(0xffff), HEX);
    if (client.connect(clientId.c_str(), mqtt_user, mqtt_pass)) {
      Serial.println("connected");
      client.subscribe("pin/control");
      client.publish("client/init", "initialized");
    } else {
      Serial.print("failed, rc=");
      Serial.print(client.state());
      Serial.println(" retrying in 5 seconds");
      delay(5000);
    }
  }
}

void setup() {
  Serial.begin(115200);
  setup_wifi();
  secureClient.setInsecure();
  client.setServer(mqtt_server, mqtt_port);
  client.setCallback(callback);
}

void loop() {
  if (!client.connected()) {
    reconnect();
  }
  client.loop();

  if (millis() - lastReadTime >= readInterval) {
    lastReadTime = millis();
    for (int pin : listeningPins) {
      int value = analogRead(pin);
      if (lastValues.find(pin) == lastValues.end() || lastValues[pin] != value) {
        lastValues[pin] = value;
        char topic[64];
        sprintf(topic, "pin/%d/data", pin);
        char payload[16];
        sprintf(payload, "%d", value);
        client.publish(topic, payload);
      }
    }
  }
}
